from aiohttp import ClientResponse


class CordedError(Exception):
    pass


## HTTP Errors


class HTTPError(CordedError):
    def __init__(self, response: ClientResponse):
        self.response = response


class BadRequest(HTTPError):
    pass


class Unauthorized(HTTPError):
    pass


class Forbidden(HTTPError):
    pass


class NotFound(HTTPError):
    pass


class PayloadTooLarge(HTTPError):
    pass


class TooManyRequests(HTTPError):
    pass


class DiscordServerError(HTTPError):
    pass



"""
MIT License

Copyright (c) 2021 vcokltfre

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from aiohttp import WSMessage, WSMsgType
from asyncio import AbstractEventLoop, sleep, Task
from sys import platform
from time import time

from .ratelimiter import Ratelimiter

from corded.objects.constants import GatewayOps, GatewayCloseCodes as CloseCodes


class Shard:
    def __init__(self, id: int, parent, loop: AbstractEventLoop):
        """A shard to connect to the Discord gateway to receive and send events.

        Args:
            id (int): The shard's ID.
            parent (corded.ws.GatewayClient): The shard's parent gateway shard manager.
            loop (AbstractEventLoop): The even loop to use.
        """

        self.id = id
        self.parent = parent
        self.loop = loop

        self.url = None
        self.ws = None

        self.session = None
        self.seq = None
        self.ws_seq = None

        self.heartbeat_task = None
        self.last_heartbeat_send = None
        self.recieved_ack = True
        self.failed_heartbeats = 0
        self.latency = None

        self.pacemaker: Task = None

        self.send_limiter = Ratelimiter(120, 60)

    async def spawn_ws(self):
        """Spawn the websocket connection to the gateway."""

        self.ws = await self.parent.http.spawn_ws(self.url)

    async def connect(self):
        """Create a connection to the Discord gateway."""

        if not self.url:
            self.url = (await self.parent.http.get_gateway()).url

        await self.spawn_ws()

        if self.session:
            await self.resume()

        await self.start_reader()

    async def close(self):
        """Gracefully close the connection."""

        if self.ws and not self.ws.closed:
            await self.ws.close()

        if self.pacemaker and not self.pacemaker.cancelled():
            self.pacemaker.cancel()

    async def send(self, data: dict):
        """Send data to the gateway.

        Args:
            data (dict): The data to send.
        """

        await self.send_limiter.wait()

        await self.parent.dispatch_send(data)
        await self.ws.send_json(data)

    async def identify(self):
        """Sends an identfy payload to the gateway."""

        await self.send({
            "op": GatewayOps.IDENTIFY,
            "d": {
                "token": self.parent.http.token,
                "properties": {
                    "$os": platform,
                    "$browser": "Corded",
                    "$device": "Corded",
                },
                "intents": self.parent.intents,
                "shard": [self.id, self.parent.shard_count],
            }
        })

    async def resume(self):
        """Resume an existing connection with the gateway."""

        await self.send({
            "op": GatewayOps.RESUME,
            "d": {
                "token": self.parent.http.token,
                "session_id": self.session,
                "seq": self.seq,
            }
        })

    async def heartbeat(self):
        """Send a heartbeat to the gateway."""

        self.last_heartbeat_send = time()

        await self.send({
            "op": GatewayOps.HEARTBEAT,
            "d": self.seq
        })

        if self.seq:
            self.seq += 1
        else:
            self.seq = 1

    async def dispatch(self, data: dict):
        """Dispatch events."""

        await self.parent.dispatch_recv(data)

        op = data["op"]

        if op == GatewayOps.HELLO:
            self.pacemaker = self.loop.create_task(self.start_pacemaker(data["d"]["heartbeat_interval"]))
            await self.identify()
        elif op == GatewayOps.ACK:
            self.latency = time() - self.last_heartbeat_send
            self.recieved_ack = True
        elif op == GatewayOps.RECONNECT:
            await self.close()
            await self.connect()

    async def handle_disconnect(self, code: int):
        """Handle the gateway disconnecting correctly."""

        if code in [
            CloseCodes.NOT_AUTHENTICATED,
            CloseCodes.AUTHENTICATION_FAILED,
            CloseCodes.INVALID_API_VERSION,
            CloseCodes.INVALID_INTENTS,
            CloseCodes.DISALLOWED_INTENTS
        ]:
            await self.parent.panic(code)

        if code in [
            CloseCodes.INVALID_SEQ,
            CloseCodes.RATE_LIMITED,
            CloseCodes.SESSION_TIMEOUT
        ]:
            self.session = None

            if code == CloseCodes.RATE_LIMITED:
                self.url = None

        self.seq = None

        await self.close()
        await self.connect()

    async def start_reader(self):
        """Start a loop constantly reading from the gateway."""

        async for message in self.ws:
            message: WSMessage

            if message.type == WSMsgType.TEXT:
                message_data = message.json()

                if s := message_data.get("s"):
                    self.ws_seq = s

                await self.dispatch(message_data)

        await self.handle_disconnect(self.ws.close_code)

    async def start_pacemaker(self, delay: float):
        """A loop to constantly heartbeat at an interval given by the gateway."""

        delay = delay / 1000

        while True:
            if not self.recieved_ack:
                self.failed_heartbeats += 1

            if self.failed_heartbeats > 2:
                await self.close()
                await self.connect()
                return

            await self.heartbeat()
            self.recieved_ack = False

            await sleep(delay)



"""
MIT License

Copyright (c) 2021 vcokltfre

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from asyncio import AbstractEventLoop, get_event_loop, sleep
from aiohttp import ClientSession, ClientResponse

from corded.constants import API_URL, VERSION
from corded.errors import HTTPError, BadRequest, Unauthorized, Forbidden, NotFound, PayloadTooLarge, TooManyRequests, DiscordServerError

from .ratelimiter import Ratelimiter
from .route import Route

import corded.objects.partials as p


class HTTPClient:
    def __init__(self, token: str, *, url: str = None, loop: AbstractEventLoop = None):
        """An HTTP client to make Discord API requests, observing ratelimits.

        Args:
            token (str): The bot token to make requests with.
            url (str, optional): The URL of the Discord API. Defaults to corded.constants.API_URL.
            loop (AbstractEventLoop, optional): The event loop to use. Defaults to asyncio.get_event_loop().
        """

        self.token = token
        self.url = url or API_URL
        self.loop = loop or get_event_loop()

        self.headers = {
            "Authorization": f"Bot {self.token}",
            "User-Agent": f"DiscordBot (Corded, https://github.com/vcokltfre/corded, version: {VERSION})",
            "X-RateLimit-Precision": "millisecond",
        }

        self.ratelimiter = Ratelimiter(self.loop)
        self.session: ClientSession = None

        self.errors = {
            "_": HTTPError,
            400: BadRequest,
            401: Unauthorized,
            403: Forbidden,
            404: NotFound,
            413: PayloadTooLarge,
        }

    @staticmethod
    async def response_as(response: ClientResponse, format: str = None):
        """Return a ClientResponse in a given format.

        Args:
            response (ClientResponse): The client response to get data from.
            format (str, optional): The format to use. Defaults to 'json', must be one of 'json', 'text', 'auto', 'raw'.
        """

        if format == "raw":
            return await response.read()
        if format == "json":
            return await response.json()
        if format == "text":
            return await response.text()
        if format == "auto":
            try:
                return await response.json()
            except:
                return await response.text()
        if format == "response":
            return response
        raise ValueError("Format must be one of 'json', 'text', 'auto', 'raw'")

    async def request(self, method: str, route: Route, *, attempts: int = None, expect: str = None, **params):
        """Make a Discord API request.

        Args:
            method (str): The HTTP method to use.
            route (Route): The Route to use for the request.
            attempts (int, optional): How many attempts to make before giving up. Defaults to 3.
            expect (str, optional): What format to expect the result in. Defaults to JSON.
        """

        attempts = attempts or 3
        expect = expect or "json"

        if not self.session or self.session.closed:
            self.session = ClientSession(headers=self.headers)

        bucket = route.bucket

        headers = {}
        if "reason" in params:
            headers["X-Audit-Log-Reason"] = params.pop("reason")

        for i in range(attempts):
            await self.ratelimiter.acquire(bucket)

            response = await self.session.request(method, self.url + route.route, **params)

            status = response.status
            headers = response.headers

            rl_reset_after = float(headers.get("X-RateLimit-Reset-After", 0))
            rl_bucket_remaining = int(headers.get("X-RateLimit-Remaining", 1))  # Default here is for non authenticated (and hence non ratelimited) endpoints
            rl_sleep_for = 0

            if status != 429 and rl_bucket_remaining == 0:
                rl_sleep_for = rl_reset_after

            if 200 <= status < 300:
                self.ratelimiter.release(bucket, rl_sleep_for)
                return await self.response_as(response, expect)

            if status == 429:
                if not headers.get("Via"):
                    raise TooManyRequests(429, response, "Ratelimited by cloudflare.")

                data = await self.response_as(response, "json")
                is_global = data.get("global", False)
                rl_sleep_for = data.get("retry_after")

                if is_global:
                    self.ratelimiter.lock_globally(rl_sleep_for)

            elif status >= 500:
                rl_sleep_for = 1 + i * 2

            else:
                self.ratelimiter.release(bucket, rl_sleep_for)
                raise self.errors.get(status, self.errors["_"])(response)

            if i == attempts - 1:
                self.ratelimiter.release(bucket, rl_sleep_for)
                continue

            await sleep(rl_sleep_for)

        self.ratelimiter.release(bucket)

        if status >= 500:
            raise DiscordServerError(response)

        raise self.errors.get(status, self.errors["_"])(response)

    async def spawn_ws(self, url: str):
        if not self.session or self.session.closed:
            self.session = ClientSession(headers=self.headers)

        args = {
            "max_msg_size": 0,
            "timeout": 60,
            "autoclose": False,
            "headers": {
                "User-Agent": self.headers["User-Agent"]
            },
        }

        return await self.session.ws_connect(url, **args)

    async def close(self):
        await self.session.close()

    async def get_gateway(self) -> p.GetGateway:
        route = Route("/gateway")
        response = await self.request("get", route)

        return p.GetGateway(**response)

    async def get_gateway_bot(self) -> p.GetGatewayBot:
        route = Route("/gateway/bot")
        response = await self.request("get", route)

        session_start_limit = p.SessionStartLimit(**response["session_start_limit"])

        return p.GetGatewayBot(response["url"], response["shards"], session_start_limit)



"""
MIT License

Copyright (c) 2021 vcokltfre

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from asyncio import AbstractEventLoop, get_event_loop
from collections import defaultdict

from .http import HTTPClient
from .ws import GatewayClient

from corded.objects import partials as p


class CordedClient:
    def __init__(self, token: str, intents: int, *, shard_ids: int = None, shard_count: int = None, loop: AbstractEventLoop = None):
        """A combined client that can make HTTP requests and connect to the gateway.

        Args:
            token (str): The token to use for API requests and connecting.
            intents (int): The intents to use while connecting to the gateway.
            loop (AbstractEventLoop, optional): The even loop to use. Defaults to asyncio.get_event_loop.
        """

        self.token = token
        self.intents = intents
        self.loop = loop or get_event_loop()

        self.http = HTTPClient(token, loop=self.loop)
        self.gateway = GatewayClient(self.http, self.intents, shard_ids, shard_count, loop=self.loop)

    def start(self):
        """Make a blocking call to start the Gateway connection."""

        self.loop.run_until_complete(self.gateway.start())

    def on(self, event: str = None):
        def wrapper(func):
            self.gateway.listeners[event].append(func)
            return func
        return wrapper



from .http import HTTPClient, Route
from .ws import GatewayClient, Shard
from .errors import (
    CordedError,
    HTTPError,
    BadRequest,
    Unauthorized,
    Forbidden,
    NotFound,
    PayloadTooLarge,
    TooManyRequests,
    DiscordServerError,
)
from .constants import VERSION as __version__
from .objects import Object
from .client import CordedClient
from .helpers import BitField

__all__ = (
    HTTPClient,
    Route,
    CordedError,
    HTTPError,
    BadRequest,
    Unauthorized,
    Forbidden,
    NotFound,
    PayloadTooLarge,
    TooManyRequests,
    DiscordServerError,
    Object,
    GatewayClient,
    Shard,
    CordedClient,
    BitField,
    __version__,
)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['corded', 'corded.http', 'corded.objects', 'corded.ws']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.7.4,<4.0.0']

setup_kwargs = {
    'name': 'corded',
    'version': '1.2.0',
    'description': 'A lightweight, extensible client library for Discord bots',
    'long_description': None,
    'author': 'vcokltfre',
    'author_email': 'vcokltfre@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from enum import Enum

AUDIT_LOG_EVENTS = {
    1: "GUILD_UPDATE",
    10: "CHANNEL_CREATE",
    11: "CHANNEL_UPDATE",
    12: "CHANNEL_DELETE",
    13: "CHANNEL_OVERWRITE_CREATE",
    14: "CHANNEL_OVERWRITE_UPDATE",
    15: "CHANNEL_OVERWRITE_DELETE",
    20: "MEMBER_KICK",
    21: "MEMBER_PRUNE",
    22: "MEMBER_BAN_ADD",
    23: "MEMBER_BAN_REMOVE",
    24: "MEMBER_UPDATE",
    25: "MEMBER_ROLE_UPDATE",
    26: "MEMBER_MOVE",
    27: "MEMBER_DISCONNECT",
    28: "BOT_ADD",
    30: "ROLE_CREATE",
    31: "ROLE_UPDATE",
    32: "ROLE_DELETE",
    40: "INVITE_CREATE",
    41: "INVITE_UPDATE",
    42: "INVITE_DELETE",
    50: "WEBHOOK_CREATE",
    51: "WEBHOOK_UPDATE",
    52: "WEBHOOK_DELETE",
    60: "EMOJI_CREATE",
    61: "EMOJI_UPDATE",
    62: "EMOJI_DELETE",
    72: "MESSAGE_DELETE",
    73: "MESSAGE_BULK_DELETE",
    74: "MESSAGE_PIN",
    75: "MESSAGE_UNPIN",
    80: "INTEGRATION_CREATE",
    81: "INTEGRATION_UPDATE",
    82: "INTEGRATION_DELETE",
}

class GatewayOps:
    DISPATCH = 0
    HEARTBEAT = 1
    IDENTIFY = 2
    PRESENCE_UPDATE = 3
    VOICE_STATE_UPDATE = 4
    RESUME = 6
    RECONNECT = 7
    REQUEST_GUILD_MEMBERS = 8
    INVALID_SESSION = 9
    HELLO = 10
    ACK = 11

class GatewayCloseCodes:
    UNKNOWN_ERROR = 4000
    UNKNOWN_OPCODE = 4001
    DECODE_ERROR = 4002
    NOT_AUTHENTICATED = 4003
    AUTHENTICATION_FAILED = 4004
    ALREADY_AUTHENTICATED = 4005
    INVALID_SEQ = 4007
    RATE_LIMITED = 4008
    SESSION_TIMEOUT = 4009
    INVALID_SHARD = 4010
    SHARDING_REQUIRED = 4011
    INVALID_API_VERSION = 4012
    INVALID_INTENTS = 4013
    DISALLOWED_INTENTS = 4014



"""
MIT License

Copyright (c) 2021 vcokltfre

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from asyncio import AbstractEventLoop, get_event_loop
from collections import defaultdict

from .shard import Shard
from .ratelimiter import Ratelimiter

from corded.objects.partials import GetGatewayBot, SessionStartLimit
from corded.helpers import int_types


class GatewayClient:
    def __init__(self, http, intents: int, shard_ids: list = None, shard_count: int = None, *, loop: AbstractEventLoop = None):
        """A client to connect to the Discord gateway.

        Args:
            http ([type]): The HTTP client to use for API requests.
            intents (int): The intents to connect with.
            shard_ids (list, optional): The shard IDs to connect with. Defaults to [0].
            shard_count (int, optional): The total number of shards being used. Defaults to 1.
            loop (AbstractEventLoop, optional): The event loop to use. Defaults to the result of asyncio.get_event_loop.
        """
        self.http = http
        self.intents = intents

        self.shard_count = shard_count or 1
        self.shard_ids = shard_ids or list(range(self.shard_count))

        self.loop = loop or get_event_loop()

        self.shards = [Shard(id, self, self.loop) for id in self.shard_ids]

        self.listeners = defaultdict(list)

    async def start(self):
        gateway: GetGatewayBot = await self.http.get_gateway_bot()
        limit: SessionStartLimit = gateway.session_start_limit

        limiter = Ratelimiter(limit.max_concurrency, 5)

        for shard in self.shards:
            await limiter.wait()
            await shard.connect()

    async def dispatch(self, event: str, raw_data: dict):
        for listener in self.listeners[event]:
            self.loop.create_task(listener(int_types(raw_data["d"])))

    async def dispatch_recv(self, data: dict):
        await self.dispatch("gateway_receive", data)

        if event := data.get("t"):
            await self.dispatch(event.lower(), data)
        else:
            await self.dispatch(f"op_{data['op']}", data)

    async def dispatch_send(self, data: dict):
        await self.dispatch("gateway_send", data)

        await self.dispatch(f"op_{data['op']}", data)
